"""Add the Suspendable class to classes that can be run/suspended/etc."""

from skytap.framework.ApiClient import ApiClient
from skytap.framework.Config import Config
import time


class Suspendable(object):

    """Add the change_state() functionality to the class.

    This includes supporting functions, like run() and suspend().
    """

    valid_states = ['running', 'suspended', 'reset', 'stopped', 'halted']

    def suspend(self, wait=False):
        """Suspend the object (environment/vm)."""
        self.change_state('suspended', wait)

    def run(self, wait=False):
        """Run the object (environment/vm)."""
        self.change_state('running', wait)

    def halt(self, wait=False):
        """Halt the object (environment/vm)."""
        self.change_state('halted', wait)

    def reset(self, wait=False):
        """Reset the object (environment/vm)."""
        self.change_state('reset', wait)

    def stop(self, wait=False):
        """Stop the object (environment/vm)."""
        self.change_state('stopped', wait)

    def change_state(self, state, wait=False):
        """Change the state of the object (environment/vm)."""
        if state not in Suspendable.valid_states:
            raise ValueError(str(state) + ' not a valid state.')

        self.refresh()
        if self.runstate == state:
            return True

        if Config.add_note_on_state_change:
            self.notes.add("Changing state via API to '" + state + "'")

        api = ApiClient()
        url = self.url + '.json'
        data = {"runstate": state}
        response = api.rest(url, {}, 'PUT', data)
        if not wait:
            return True

        if state == 'reset':
            state = 'running'

        if state == 'halted':
            state = 'stopped'

        self.refresh()
        counter = 0
        while not self.runstate == state and counter < 12:
            time.sleep(10)
            self.refresh()
        if self.runstate == state:
            return True
        return False



"""Skytap API object wrapping Skytap templates.

This roughly translates to the Skytap API call of /v2/templates REST call,
but gives us better access to the bits and pieces of the templates.

If accessed via the command line (``python -m skytap.Templates``) this will
return the templates from Skytap in a JSON format.
"""
import sys

from skytap.models.SkytapGroup import SkytapGroup
from skytap.models.Template import Template


class Templates(SkytapGroup):

    """Set of Skytap templates.

    Example:
        t = skytap.Templates()
        print len(t)
    """

    def __init__(self):
        """Build an initial list of templates."""
        super(Templates, self).__init__()
        self.load_list_from_api('/v2/templates', Template)

    def vm_count(self):
        """Count the total number of VMs."""
        count = 0
        for e in self.data:
            count += self.data[e].vm_count
        return count

    def svms(self):
        """Count the total number of SVMs in use."""
        count = 0
        for e in self.data:
            count += self.data[e].svms
        return count

    def storage(self):
        """Count the total amount of storage in use."""
        count = 0
        for e in self.data:
            count += self.data[e].storage
        return count

if __name__ == '__main__':
    print(Templates().main(sys.argv[1:]))



"""Functions needed to access the skytap api."""
import json
import requests
import six
import sys
import time

from skytap.framework.Config import Config
import skytap.framework.Utils as Utils

requests.packages.urllib3.disable_warnings()


class ApiClient(object):

    """Wrap the calls to the Skytap API."""

    def __init__(self):
        """Initial setup of things.

        Also does some basic sanity checking on the config to make sure we
        have what we need to be able to access the skytap API.
        """
        super(ApiClient, self).__init__()

        if not Config.base_url:
            raise ValueError('Invalid base_url')

        if not Config.user:
            raise ValueError('Invalid api_user')

        if not Config.token:
            raise ValueError('Invalid api_token')

        self.auth = (Config.user, Config.token)

        self.last_headers = None
        self.last_status = 0
        self.last_range = 0

        self.cmds = {
            'GET': requests.get,
            'PUT': requests.put,
            'POST': requests.post,
            'DELETE': requests.delete
        }

        self.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

    def _check_response(self, resp, attempts=1):
        """Return true if the rseponse a good/reasonable one.

        If the HTTP status code is in the 200s, return True,
        otherwise try to determine what happened. If we're asked to retry,
        politely wait the appropraite amount of time and retry, otherwise,
        wait the retry_wait amount of time.

        Fail (return False) if we've exceeded our retry amount.
        """
        if resp is None:
            raise ValueError('A response wasn\'t received')

        if 200 <= resp.status_code < 300:
            return True

        # If we made it this far, we need to handle an exception
        if attempts >= Config.max_http_attempts or (resp.status_code != 429 and
                                                    resp.status_code != 423):
            raise Exception(json.loads(resp.text))

        if resp.status_code == 423:  # "Busy"
            if 'Retry-After' in resp.headers:
                Utils.info('Received HTTP 423. Retry-After set to ' +
                              resp.headers['Retry-After'] + ' sec. Waiting to retry.')  # noqa
                time.sleep(int(resp.headers['Retry-After']) + 1)
            else:
                Utils.info('Received HTTP 429. Too many requests. Waiting to retry.')  # noqa
                time.sleep(Config.retry_wait)
            return False

        # Assume we're going to retry with exponential backoff
        # Should only get here on a 429 "too many requests" but it's
        # not clear from Skytap what their limits are on when we should retry.
        time.sleep(2 ** (attempts - 1))

        return False

    @staticmethod
    def _dict_to_query_params(d):
        """Return proper query string to add to a url.

        Turns {'count': 5, 'offset': 2} into '?count=5&offset=2'.
        """
        if d is None or len(d) == 0:
            return ''

        param_list = [param + '=' +
                      (str(value).lower()
                       if type(value) == bool else str(value))
                      for param, value in six.iteritems(d)
                      if value is not None]
        return '?' + "&".join(param_list)

    def rest(self, url, params=None, req='get', data=None):
        """Call the REST API, returning all results.

        This calls the actual REST API, then checks the returning headers in
        case there is a range returned, implying that the full range wasn't
        returned originally. If there's a range, then make a second call asking
        for everything.

        This defeats the pagination that Skytap uses in their v2 API, but is
        useful for us given how we use the API.
        """
        if params is None:
            params = {}
        if not url.upper().startswith('HTTP'):
            url = Config.base_url + url

        first_call = self._rest(req, url, params, data)
        if self.last_range == 0:
            return first_call

        params['offset'] = 0
        params['count'] = self.last_range

        return self._rest(req, url, params, data)

    def _rest(self, req, url, params=None, data=None, attempts=0):
        """Send a rest rest request to the server."""
        if params is None:
            params = {}
        cmd = req.upper()
        if cmd not in self.cmds.keys():
            raise ValueError("Command type (" + cmd + ") not recognized.")

        url += self._dict_to_query_params(params)

        response = self.cmds[cmd](url,
                                  headers=self.headers,
                                  auth=self.auth,
                                  params=data)

        self.last_status = response.status_code
        self.last_headers = response.headers
        self.last_range = 0
        if "content-range" in self.last_headers:
            self.last_range = self.last_headers["content-range"].split("/")[1]

        attempts += 1
        if self._check_response(response, attempts):
            try:
                return json.dumps(json.loads(response.text), indent=4)
            except ValueError:
                return response.text
        else:
            return self._rest(req, url, params, data, attempts)



"""Base object to handle groups of Skytap objects."""
import argparse
import json
import six

from skytap.framework.ApiClient import ApiClient
from skytap.framework.Json import SkytapJsonEncoder


class SkytapGroup(ApiClient, six.Iterator):

    """Base object for use with Skytap resource groups."""

    def __init__(self):
        super(SkytapGroup, self).__init__()
        self.data = {}
        self.itercount = 0
        self.search_fields = ['name']

    def load_list_from_api(self, url, target, params=None):
        """Load something from the Skytap API and fill this object.

        .. note: This should rarely be called by anything but a child object,
            typically in its __init__() method.

        Args:
            url (str): The Skytap URL to load ('/v2/users').
            target: The :class:`~skytap.models.SkytapResource`
                type to load (For example: 'User')
            params (dict): Any URL parameters to add to URL.

        This should look like, in the child object::

            self.load_list_from_api('/v2/projects', Project)

        """
        if params is None:
            params = {}
        self.load_list_from_json(self.rest(url, params), target, url)
        self.params = params

    def load_list_from_json(self, json_list, target, url=None):
        """Load items from a json list and fill this object.

        Args:
            json_list (list): The list to load the items from.
            target: The :class:`~skytap.models.SkytapResource`
                type to load (for example: 'User')

        This should look like, in the child object:

        .. code-block: python

            self.load_list_from_json(json, Project)

        """
        self.data = {}
        self.target = target

        if url is not None:
            self.url = url
            if '/v2/' in self.url:
                self.url_v1 = self.url.replace('/v2/', '/')
                self.url_v2 = self.url
            else:
                self.url_v1 = self.url
                self.url_v2 = None

        if isinstance(json_list, str):
            self.json_from_load = json.loads(json_list)
        elif isinstance(json_list, list):
            self.json_from_load = json_list
        else:
            raise TypeError
        for j in self.json_from_load:
            # prefer an int for the ID since that's the most common and
            # easiest to work with, but some things (quotas) have string
            # ids, so we want to leave those alone. A given resource/group
            # will have to handle non-int keys differently if there's a
            # case where that's a problem.
            try:
                self.data[int(j['id'])] = target(j)
            except ValueError:
                self.data[j['id']] = target(j)

    def first(self):
        """Return the first record in the list.

        Mainly used to get a single arbitrary object for testing.

        .. note: The selection of this list depends on how the objects
            are returned from Skytap. This ordering should be considered
            arbitrary and not reliable from one run to the next.

        Returns:
            ~skytap.models.SkytapResource: An object from the list.
        """
        return self.data[list(self.data)[0]]

    def find(self, search):
        """Return a list of objects, based on the search criteria.

        This looks for matching ids if the search is a number, or searches
        the name if search is a string.

        Args:
            search (int or str): What to search for.
        Returns:
            List: Any environments matching the search criteria.

        Example:
            >>> envs = skytap.Environments().search('testing')

        """
        found = []
        for one in list(self.data):
            test = self.data[one]
            try:
                if int(search) == test.id:
                    found.append(test)
            except ValueError:
                if search == str(test.id):
                    found.append(test)
                pass
            for field in self.search_fields:
                if str(search).upper() in str(test.data[field]).upper():
                    found.append(test)
        return found

    def refresh(self):
        """Reload our data."""
        self.load_list_from_api(self.url,
                                self.target,
                                self.params)

    def main(self, argv):
        """What to do when called from the command line.

        This function is usually accessed via the command line::

            python -m skytap.Environments

        but can be used to return quick sets of formatted JSON::

            >>> print(skytap.Environments().main())

        Anything passed to the function will be searched for::

            python -m skytap.Users fozzy

        and::

            >>> print(skytap.Users().main('scooter'))

        Args:
            argv (list): Command line arguments

        Returns:
            str: Formatted JSON of the request.
        """
        obj_type = type(self.data[list(self.data)[0]])
        obj_name = obj_type.__name__
        obj_id_type = type(self.data[list(self.data)[0]].id)

        if len(argv) > 0:
            found = self.find(argv[0])
            json_return = []
            for item in found:
                json_return.append(json.loads(item.json()))
            return json.dumps(json_return, indent=4)

        return json.dumps(self.json(), indent=4)

    def json(self):
        """Convert our list into a json."""
        json_return = []
        for item in self.data:
            json_return.append(json.loads(self.data[item].json()))
        return json_return

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return json.dumps(self.json, indent=4)

    def __getitem__(self, key):
        return self.data[key]

    def __iter__(self):
        return self

    def __next__(self):
        if self.itercount >= len(self.data):
            raise StopIteration
        n = list(self.data)[self.itercount]
        self.itercount += 1
        return self.data[n]

    def keys(self):
        """Return the keys from the group list."""
        return self.data.keys()

    def __contains__(self, key):
        return key in self.data



"""Support for Skytap API access to the company quotas.

If accessed via the command line (``python -m skytap.Quotas``) this will
return the quotas from Skytap in a JSON format.
"""
import sys

from skytap.models.Quota import Quota
from skytap.models.SkytapGroup import SkytapGroup


class Quotas(SkytapGroup):

    """Company/account quotas object."""

    def __init__(self):
        """Load the quotas from Skytap."""
        super(Quotas, self).__init__()
        self.load_list_from_api('/v2/company/quotas', Quota)

if __name__ == '__main__':
    print(Quotas().main(sys.argv[1:]))



"""Better support to convert our objects to JSON."""
from datetime import datetime
import json


class SkytapJsonEncoder(json.JSONEncoder):

    """Convert a few additional types into JSON."""

    def default(self, o):
        """Convert datetime and our custom objects to JSON.

        Our objects all have a .json() function available to them,
        so we can just try using that if it's available.
        """
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)

        if (isinstance(o, datetime)):
            return o.isoformat()

        try:
            ret = json.loads(o.json())
        except AttributeError:
            raise
        else:
            return ret

        return json.JSONEncoder.default(self, o)



"""Support for Skytap interfacess."""
import json

from skytap.framework.ApiClient import ApiClient  # noqa
from skytap.models.Interface import Interface  # noqa
from skytap.models.PublishedServices import PublishedServices  # noqa
from skytap.models.SkytapGroup import SkytapGroup  # noqa


class Interfaces(SkytapGroup):
    """A list of Interfaces."""

    def __init__(self, interface_json, vm_url):
        """Create the list of Interfaces.

        Args:
            interfaces_json (string): The JSON from Skytap API to build the list
                                      from.
        """
        super(Interfaces, self).__init__()
        self.load_list_from_json(interface_json, Interface)
        for i in self.data:
            self.data[i].data['url'] = (vm_url + "/interfaces/"
                                        "" + self.data[i].id)

            api = ApiClient()
            services_json = json.loads(api.rest(self.data[i].url))
            self.data[i].services = PublishedServices(services_json["services"],
                                                      self.data[i].url)



from skytap.Environments import Environments  # noqa
from skytap.Groups import Groups  # noqa
from skytap.Projects import Projects  # noqa
from skytap.Quotas import Quotas  # noqa
from skytap.Templates import Templates  # noqa
from skytap.Users import Users  # noqa
from skytap.Vpns import Vpns  # noqa



"""Skytap API object wrapping Skytap VPNs.

This roughly translates to the Skytap API call of /v2/vpns REST call,
but gives us better access to the bits and pieces of the VPN.

If accessed via the command line (``python -m skytap.Vpns``) this will
return the quotas from Skytap in a JSON format.
"""
import sys

from skytap.models.SkytapGroup import SkytapGroup
from skytap.models.Vpn import Vpn


class Vpns(SkytapGroup):

    """Set of Skytap VPNs.

    Example:
        v = skytap.Users()
        print len(v)
    """

    def __init__(self):
        """Build the VPN list from the Skytap API."""
        super(Vpns, self).__init__()
        self.load_list_from_api('/v2/vpns', Vpn)

if __name__ == '__main__':
    print(Vpns().main(sys.argv[1:]))



"""Support for Skytap quotas."""
from datetime import timedelta

from skytap.models.SkytapResource import SkytapResource


class Quota(SkytapResource):

    """One piece of quota information."""

    def __init__(self, quota_json):
        """Build the quota object.

        Args:
            quota_json (list): The quota data.
        """
        super(Quota, self).__init__(quota_json)

    def _calculate_custom_data(self):
        """Create a percentage used and time object, if applicable."""
        if self.limit is not None:
            self.data['pct'] = self.usage / self.limit
        if self.units == 'hours':
            self.time = timedelta(hours=self.usage)
        self.data['name'] = self.id

    def __str__(self):
        quota = self.id + " = " + str(self.usage)
        if self.units != 'integer':
            quota += "" + self.units
        return quota



"""Skytap API object wrapping Skytap Environments.

This roughly translates to the Skytap API call of /v2/configurations REST call,
but gives us better access to the bits and pieces of the environments.

**Accessing via command line**

If accessed via the command line this will return the environments from
Skytap in a JSON format::

    python -m skytap.Environments

If you know the environment you want information on, you can also specify
it directly. You can search by id or by a part of the environment name::

    python -m skytap.Environments 12345
    python -m skytap.Environments test

Additionally, you can search on some other crieria of a group to get a set
you're looking for.

Runstate::

    python -m skytap.Environments running  # or 'suspended',
        'stopped', or 'busy'

Region::

    python -m skytap.Environments us-west

**Accessing via Python**

After you've installed ``skytap`` and added ``import skytap`` to your
script, you can access the Skytap environments by the
:class:`skytap.Environments` object.

Example:

.. code-block:: python

    envs = skytap.Environments()
    for e in envs:
        print (e.name)

Each environment has many things you can do with it - see the
:class:`skytap.models.Environment` object for actions you can take on an
individual environment.

On the full list of environments, you can also get a vm count, svm count,
get global storage, and delete environments. Each action is documented, below.

Environments can also perform any of the actions of other
:class:`SkytapGroup` objects. See the documentation
on the :class:`skytap.models.SkytapGroup` class for
information there.

.. note::
    Some pieces of a given environment, specifically `notes` and `user_data`,
    are only available via additional calls to the API. These fields will
    not exist when first creating the environments object, but any direct
    access to those fields will trigger the API call behind the scenes.

    This is important if you're listing the entire contents (say, sending it
    to a JSON) - these fields won't be included if you haven't made that direct
    access.

    This is by design to conserve API calls as most usage doesn't need or use
    those fields.
"""
import sys

from skytap.models.Environment import Environment
from skytap.models.SkytapGroup import SkytapGroup


class Environments(SkytapGroup):

    """Set of Skytap environments."""

    def __init__(self):
        """Build an initial list of environments."""
        super(Environments, self).__init__()
        self.load_list_from_api('/v2/configurations',
                                Environment,
                                {'scope': 'company'})
        self.search_fields.append('runstate')
        self.search_fields.append('region')

    def vm_count(self):
        """Count the total number of VMs.

        Returns:
            int: Number of VMs used across all environments.

        Example:

        .. code-block:: python

            envs = skytap.Environments()
            print(envs.vm_count())
        """
        count = 0
        for e in self.data:
            count += self.data[e].vm_count
        return count

    def svms(self):
        """Count the total number of SVMs in use.

        Returns:
            int: Number of SVMs used across all environments.

        Example:

        .. code-block:: python

            envs = skytap.Environments()
            print(envs.svms())
        """
        count = 0
        for e in self.data:
            count += self.data[e].svms
        return count

    def storage(self):
        """Count the total amount of storage in use.

        Returns:
            int: Amount of storage used across all environments.

        Example:

        .. code-block:: python

            envs = skytap.Environments()
            print(envs.storage()))
        """
        count = 0
        for e in list(self.data):
            count += self.data[e].storage
        return count

    def delete(self, env):
        """Delete a given environment.

        .. warning::
            This is unrecoverable. Use with **extreme** caution.

        Args:
            env: The :class:`~skytap.models.Environment` to delete.

        Returns:
            bool: True if the environment was deleted.
        Raises:
            KeyError: If ``env`` isn't in the Environments set.

        Example:

        .. code-block:: python

            envs = skytap.Environments()
            target = envs[12345]
            envs.delete(target)
        """
        target_id = env.id
        if isinstance(env, Environment):
            if target_id not in self.data:
                raise KeyError
            env.delete()
        elif isinstance(env, int):
            if env not in self.data:
                raise KeyError
            self.data[env].delete()
        else:
            raise KeyError
        self.refresh()
        return target_id not in self.data

if __name__ == '__main__':
    print(Environments().main(sys.argv[1:]))



"""Base class for all Skytap Resources."""
import json

from skytap.framework.ApiClient import ApiClient  # noqa
from skytap.framework.Json import SkytapJsonEncoder  # noqa
import skytap.framework.Utils as Utils  # noqa


class SkytapResource(object):

    """Represents one Skytap Resource - a VM, Environment, User, whatever."""

    def __init__(self, initial_json):
        super(SkytapResource, self).__init__()

        self.data = {}
        self.data["id"] = 0
        for k in initial_json.keys():
            self.data[k] = initial_json[k]
        if 'url' in self.data:
            if '/v2/' in self.url:
                self.data['url_v1'] = self.url.replace('/v2/', '/')
                self.data['url_v2'] = self.url
            else:
                self.data['url_v1'] = self.url
                self.data['url_v2'] = None
        self._convert_data_elements()
        self._calculate_custom_data()

    def _calculate_custom_data(self):
        """Used so objects can create and calculate new data elements."""
        pass

    def _convert_data_elements(self):
        """Convert some data elements into variable types that make sense."""

        try:
            self.data['created_at'] = Utils.convert_date(self.data['created_at'])  # noqa
        except (ValueError, AttributeError, KeyError):
            pass

        try:
            self.data['last_login'] = Utils.convert_date(self.data['last_login'])  # noqa
        except (ValueError, AttributeError, KeyError):
            pass

        try:
            self.data['last_run'] = Utils.convert_date(self.data['last_run'])
        except (ValueError, AttributeError, KeyError):
            pass

        try:
            self.data['updated_at'] = Utils.convert_date(self.data['updated_at'])  # noqa
        except (ValueError, AttributeError, KeyError):
            pass
        try:
            self.data['last_installed'] = Utils.convert_date(self.data['last_installed'])  # noqa
        except (ValueError, AttributeError, KeyError):
            pass

        try:
            self.data['id'] = int(self.data['id'])
        except (ValueError, AttributeError, KeyError):
            pass

    def refresh(self):
        """Refresh the data in our object, if we have a URL to pull from."""
        if 'url' not in self.data:
            return KeyError
        api = ApiClient()
        env_json = api.rest(self.url)
        self.__init__(json.loads(env_json))

    def __getattr__(self, key):
        if key not in self.data:
            raise AttributeError
        return self.data[key]

    def json(self):
        """Convert the object to JSON."""
        return json.dumps(self.data, indent=4, cls=SkytapJsonEncoder)

    def __str__(self):
        return str(self.name)

    def __int__(self):
        return int(self.id)

    def __gt__(self, other):
        return int(self) > int(other)

    def __lt__(self, other):
        return int(self) < int(other)

    def __hash__(self):
        return hash(repr(self.data))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __contains__(self, key):
        return key in self.data



"""Support for Skytap services."""
import json

from skytap.models.SkytapGroup import SkytapGroup  # noqa
from skytap.models.PublishedService import PublishedService  # noqa


class PublishedServices(SkytapGroup):
    """A list of Published Services."""

    def __init__(self, service_json, interface_url):
        """Create the list of Published Services.

        Args:
            services_json (string): The JSON from Skytap API to build the list
                                    from.
        """
        super(PublishedServices, self).__init__()
        self.load_list_from_json(service_json, PublishedService)
        for s in self.data:
            self.data[s].data["url"] = (interface_url + "/services/"
                                        "" + str(self.data[s].id))



"""Skytap API object wrapping Skytap Groups.

This roughly translates to the Skytap API call of /v2/groups REST call,
but gives us better access to the bits and pieces of the groups.

**Accessing via command line**

If accessed via the command line this will return the environments from
Skytap in a JSON format::

    python -m skytap.Groups

If you know the environment you want information on, you can also specify
it directly. You can search by id or by a part of the environment name::

    python -m skytap.Groups 12345
    python -m skytap.Groups test


**Accessing via Python**

You can access the Skytap environments by the :class:`skytap.Groups` object.

Example:

.. code-block:: python

    groups = skytap.Groups()
    for g in groups:
        print(g.name)

Each group has many things you can do with it - see the
:class:`skytap.models.Group` object for actions you can take on an
individual group.

On the full list of groups, you can also do a few other things:

    - :func:`add`: add a new group.
    - :func:`delete`: delete a group.

Environments can also perform any of the actions of other
:class:`SkytapGroup` objects. See the documentation
on the :class:`skytap.models.SkytapGroup` class for
information there.
"""
import json
import sys

from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.Group import Group
from skytap.models.SkytapGroup import SkytapGroup


class Groups(SkytapGroup):

    """Set of Skytap groups.

    Generally, access this through simply creating a
    :class:`skytap.Groups` object.

    Example:

    .. code-block:: python

        groups = skytap.Groups()
        for g in groups:
            print(g.name)
    """

    def __init__(self, json_list=None):
        """Build an initial list of groups.

        If the first parameter is missing, go to the API to get the full list.
        """
        super(Groups, self).__init__()
        if json_list is None:
            self.load_list_from_api('/v2/groups', Group)
        else:
            self.load_list_from_json(json_list, Group)

    def add(self, group, description=''):
        """Add one group.

        Args:
            group (str): The group name to add.
            description (str): The group description to add.

        Returns:
            int: The new group id from Skytap.

        Example:

        .. code-block:: python

            groups = skytap.Groups()
            new_group = groups.add('muppets', 'felt covered friends')
            print(groups[new_group].name)
        """
        Utils.info('Adding group: ' + group)
        api = ApiClient()
        data = {"name": group,
                "description": description}
        url = self.url + '.json'
        response = api.rest(url, data, 'POST')
        new_group = json.loads(response)
        self.refresh()
        if 'id' in new_group:
            return int(new_group['id'])
        Utils.warning('Trying to create group (' + group + '), but ' +
                      'got an unexpected return from Skytap. Response:\n' +
                      response)
        return 0

    def delete(self, group):
        """Delete a group.

        .. warning::
            This is unrecoverable. Use with caution.

        Args:
            group: The :class:`~skytap.models.Group` to delete.

        Returns:
            bool: True if group deleted.

        Raises:
            TypeError: if group is not a :class:`~skytap.models.Group`
        """
        if isinstance(group, int):
            if group not in self.data:
                raise KeyError
            group = self.data[group]
        if not isinstance(group, Group):
            raise KeyError

        target_id = group.id

        group.delete()
        self.refresh()

        return target_id not in self.data

if __name__ == '__main__':
    print(Groups().main(sys.argv[1:]))



"""Support for Skytap groups."""
import json

from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.SkytapResource import SkytapResource
from skytap.Users import Users


class Group(SkytapResource):

    def __getattr__(self, key):
        """Load values for anything that doesn't get loaded by default.

        For user_data and notes, a secondary API call is needed. Only make that
        call when the info is requested.
        """
        if key == 'users':
            if key in self.data:
                return self.data[key]
            api = ApiClient()
            user_json = api.rest(self.url)
            self.data['users'] = Users(json.loads(user_json)['users'])
            return self.users

        return super(Group, self).__getattr__(key)

    def _calculate_custom_data(self):
        if 'users' in self.data:
            if (isinstance(self.data['users'], list)):
                if len(self.data['users']) > 0:
                    if isinstance(self.data['users'][0], dict):
                        self.data['users'] = Users(self.data['users'])

    def remove_user(self, user):
        """Remove a :class:`user` from the group.

        Args:
            user (int): id of the user to remove.

        Raises:
            TypeError: If user is not an :class:`int`.
            KeyError: If user is not in :class:`Users` list.

        Returns:
            bool: True if the user was removed.

        Example:
            >>> groups = skytap.Groups()
            >>> groups[1234].remove_user(12345)
        """
        if (type(user) is not int):
            raise TypeError('User must be an int.')

        Utils.info('Removing user ' + str(user) +
                   ' from group: ' + self.name)
        api = ApiClient()
        user_json = api.rest(self.url + '/users/' + str(user),
                             {},
                             'DELETE')

        self.refresh()
        return user not in self.users

    def add_user(self, user):
        """Add a :class:`User` to the group.

        Args:
            user (int): id of the user to add.

        Raises:
            TypeError: If user is not an :class:`int`.
            KeyError: If user is not in :class:`Users` list.

        Returns:
            bool: True if the user was added.

        Example:
            >>> groups = skytap.Groups()
            >>> users = skytap.Users()
            >>> for u in users:
            ...     groups[12345].add(u.id)
        """
        if (type(user) is not int):
            raise TypeError('User must be an int.')

        Utils.info('Adding user ' + str(user) + ' to group: ' + self.name)
        api = ApiClient()
        user_json = api.rest(self.url + '/users/' + str(user) + '.json',
                             {},
                             'PUT')
        self.refresh()
        return user in self.users

    def delete(self):
        """Delete the group."""
        Utils.info('Deleting group: ' +
                   str(self.id) + ' (' + self.name + ')')
        api = ApiClient()
        response = api.rest(self.url_v1,
                            {},
                            'DELETE')
        return response



import sys

from skytap.models.Project import Project
from skytap.models.SkytapGroup import SkytapGroup


class Projects(SkytapGroup):
    def __init__(self):
        super(Projects, self).__init__()
        self.load_list_from_api('/v2/projects', Project)

if __name__ == '__main__':
    print(Projects().main(sys.argv[1:]))



"""Support for a single note in a Skytap environment or vm."""
from skytap.models.SkytapResource import SkytapResource


class Note(SkytapResource):

    """One note."""

    def __init__(self, note_json):
        super(Note, self).__init__(note_json)

    def __str__(self):
        return self.text



from skytap.Environments import Environments
from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.SkytapResource import SkytapResource


class User(SkytapResource):
    def __init__(self, user_json):
        super(User, self).__init__(user_json)

    def _calculate_custom_data(self):
        if 'sso_enabled' in self.data:
            self.data['sso'] = bool(self.sso_enabled)
        else:
            self.data['sso'] = False

        self.data['name'] = self.first_name + ' ' + self.last_name
        if 'account_role' in self.data:
            self.data['admin'] = self.account_role == 'admin'
        if 'configurations' in self.data:
            if (isinstance(self.data['configurations'], list)):
                if len(self.data['configurations']) > 0:
                    if isinstance(self.data['configurations'][0], dict):
                        self.data['configurations'] = Environments(self.data['configurations'])  # noqa

    def delete(self, transfer_user):
        """Delete the user."""
        if isinstance(transfer_user, User):
            transfer_user = transfer_user.id
        if not isinstance(transfer_user, int):
            raise TypeError('transfer_user must be a User or int.')
        Utils.info('Deleting user: ' +
                   str(self.id) + ' (' + self.name + ') and transferring ' +
                   'resources to user id: ' + str(transfer_user))
        api = ApiClient()

        transfer = {"transfer_user_id": str(transfer_user)}

        response = api.rest(self.url,
                            transfer,
                            'DELETE')
        return response



"""Support for Skytap VMs."""
import json

from skytap.models.SkytapGroup import SkytapGroup
from skytap.models.Vm import Vm


class Vms(SkytapGroup):

    """A list of VMs."""

    def __init__(self, vms_json, env):
        """Create the list of VMs.

        Args:
            vms_json (string): The JSON from Skytap API to build the list from.
            parent (Environment or Template): The parent object -
                                              environment or template.

        """
        super(Vms, self).__init__()
        self.load_list_from_json(vms_json, Vm)
        for v in self.data:
            self.data[v].data['url'] = ('/v2/configurations/' + str(env) +
                                        '/vms/' + str(self.data[v].id))



"""Support for a published service resource in Skytap."""
from skytap.framework.ApiClient import ApiClient  # noqa
import skytap.framework.Utils as Utils  # noqa
from skytap.models.SkytapResource import SkytapResource  # noqa


class PublishedService(SkytapResource):

    def delete(self):
        """Delete a service. Cannot be undone!"""
        Utils.info('Deleting published service: ' + str(self.id))
        api = ApiClient()
        response = api.rest(self.url, {}, 'DELETE')
        return response



"""Handle the config file and such for the Skytap system."""
import json
import os
import six

import skytap.framework.Utils as Utils

initial_config = {'user': '',           # Should only be defined in env vars.
                  'token': '',          # Should only be defined in env vars.
                  'log_level': 30,      # 0 - 50. Debug=10, Info=20, Warn=30
                  'base_url': 'https://cloud.skytap.com',
                  'max_http_attempts': 4,
                  'retry_wait': 10,     # Skytap recommends waiting 10 sec.
                  'add_note_on_state_change': True
                  }


class ConfigType(type):

    """A meta class for Config.

    This allows magic methods to be used on the Config class, making things
    like 'Config.token' work as well as len(Config) even when Config is a class
    and not an object. This makes for cleaner use in other classes.
    """

    def __getattr__(cls, key):
        """Make the config values accessible.

        This allows all config values to be available via calls like:
        Config.user
        """
        if key not in cls.config_data:
            # These are called during nose setup before logging is turned off
            # during testing. Not the best, but tests look better with these
            # supressed.
            if key not in ['__test__', 'address', 'im_class', '__self__']:
                Utils.error("Tried to access config value '" +
                            str(key) + "', which doesn't exist.")
            raise AttributeError
        return cls.config_data[key]

    def __len__(cls):
        """Expose how many config items we have."""
        return len(cls.config_data)

    def __str__(cls):
        """A string representation of the config, JSON formatted, and prettified.

        Token is excluded from this, so this can be safely printed for
        debugging.
        """
        temp_config = cls.config_data.copy()
        temp_config["token"] = ''
        return json.dumps(temp_config, indent=4)

    def __repr__(cls):
        """A string representation of the config, JSON formatted.

        Token is excluded from this, so this can be safely printed for
        debugging.
        """
        temp_config = cls.config_data.copy()
        temp_config["token"] = ''
        return json.dumps(temp_config)

    def __dir__(cls):
        """List only items in the config_data list.

        Polite since we're implementing __getattr__.
        """
        dir_list = []
        for config_item in cls.config_data:
            dir_list.append(config_item)
        return dir_list

    def __contains__(cls, item):
        """Allow checks for items in the config list."""
        return item in cls.config_data

    def __iter__(cls):
        """Allow 'for x in config'.

        Ultimately, this passes the 'how to iterate' problem down to the
        config_data object and lets that object handle the actual iteration.
        """
        return iter(cls.config_data)


@six.add_metaclass(ConfigType)
class Config(object):

    """Contain all of our config values into this object."""

    # Some very basic defaults. This will allow the module to work by
    # just creating the SKYTAP_TOKEN and SKYTAP_USER env variables
    # for most things you'd want to do.
    config_data = initial_config

# Load config values and set up the class.

for key in Config:
    env_val = "SKYTAP_" + key.upper()
    if env_val in os.environ:
        Config.config_data[key] = os.environ[env_val]
        try:
            Config.config_data[key] = int(Config.config_data[key])
        except ValueError:
            pass

if os.environ.get('READTHEDOCS', None) != 'True':
    if Config.base_url != 'https://cloud.skytap.com':
        Utils.warning('Base URL is not Skytap\'s recommended value. ' +
                      'This very likely will break things.')

    if len(Config.token) == 0:
        Utils.error('No environment variable SKYTAP_TOKEN found. ' +
                    'Set this variable and try again.')
        raise ValueError

    if len(Config.user) == 0:
        Utils.error('No environment variable SKYTAP_USER found. ' +
                    'Set this variable and try again.')
        raise ValueError

Utils.log_level(Config.log_level)



"""Support for notes that are attached to VMs and environments."""
from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.Note import Note
from skytap.models.SkytapGroup import SkytapGroup


class Notes(SkytapGroup):

    """A collection of notes."""

    def __init__(self, note_json, env_url):
        """Build note list."""
        super(Notes, self).__init__()
        self.load_list_from_json(note_json, Note)
        self.url = env_url + '/notes.json'

    def add(self, note):
        """Add one note.

        Args:
            note (str): The note text to add.

        Returns:
            str: The response from Skytap, typically the new note.
        """
        Utils.info('Adding note: ' + note)
        api = ApiClient()
        data = {"text": note}
        response = api.rest(self.url, data, 'POST')
        self.refresh()
        return response

    def delete(self, note):
        """Delete one note.

        Args:
            note: The :class:`~skytap.models.Note` to delete.

        Returns:
            str: The response from Skytap.

        Raises:
            TypeError: If note is not a Note object.
        """
        if note is None:
            return False
        if not isinstance(note, Note):
            raise TypeError
        Utils.info('Deleting note ID: ' + str(note.id))
        api = ApiClient()
        url = self.url.replace('.json', '/' + str(note.id))
        response = api.rest(url,
                            {},
                            'DELETE')
        self.refresh()
        return response

    def oldest(self):
        """Return the oldest note.

        Returns:
            ~skytap.models.Note: The oldest note.

        Used most often to delete the oldest note.

        Example:

        .. code-block:: python

            notes = skytap.Environtments().first.notes
            print(notes.oldest().text)
            # notes.delete(notes.oldest())  # most common use case.
        """
        target = None
        for n in self.data:
            if target is None:
                target = self.data[n]
                continue
            if self.data[n].updated_at > target.updated_at:
                target = self.data[n]
        return target

    def newest(self):
        """Return the newest note.

        Returns:
            ~skytap.models.Note: The newest note.
        """
        target = None
        for n in self.data:
            if target is None:
                target = self.data[n]
                continue
            if self.data[n].updated_at < target.updated_at:
                target = self.data[n]
        return target

    def delete_all(self):
        """Delete all notes.

        Returns:
            int: count of deleted notes.

        Use with care!
        """
        Utils.debug('Deleting all notes.')
        keys = self.data.keys()
        count = len(keys)
        for key in keys:
            self.delete(self.data[key])
        self.refresh()
        return count

    def refresh(self):
        """Refresh the notes.

        Raises:
            KeyError: if the Notes object doesn't
                have a url attribute for some reason.

        Go back to Skytap and get the notes again. Useful when you've changed
        the notes and to make sure you're current.
        """
        if len(self.url) == 0:
            return KeyError
        api = ApiClient()
        note_json = api.rest(self.url)
        self.load_list_from_json(note_json, Note)



"""Skytap API object wrapping Skytap users.

This roughly translates to the Skytap API call of /v2/users REST call,
but gives us better access to the bits and pieces of the user.

If accessed via the command line (``python -m skytap.Users``) this will
return the users from Skytap in a JSON format.
"""
import json
import sys

from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.SkytapGroup import SkytapGroup
from skytap.models.User import User


class Users(SkytapGroup):

    """Set of Skytap users.

    Example:

    .. code-block:: python
        u = skytap.Users()
        print len(u)
    """

    def __init__(self, json_list=None):
        """Build an initial list of users.

        If the first parameter is missing, go to the API to get the full list.
        """
        super(Users, self).__init__()
        if json_list is None:
            self.load_list_from_api('/v2/users', User)
        else:
            self.load_list_from_json(json_list, User)

    def admins(self):
        """Count the numbers of admins."""
        count = 0
        for u in self.data:
            if self[u].admin:
                count += 1
        return count

    def add(self, login_name, email=None):
        """Add one user.

        Args:
            login_name (str): The login id of the account, usually an email.
            email (str): The email of the account. If blank, will use
                login_name.

        Returns:
            int: The new user id from Skytap.

        Example:

        .. code-block:: python

            users = skytap.Users()
            new_user = users.add('kermit.frog@fulcrum.net')
            print(users[new_user].login_name)
        """
        Utils.info('Adding user: ' + login_name)
        if email is None:
            email = login_name
        api = ApiClient()
        data = {"login_name": login_name,
                "email": email}
        url = self.url_v1 + '.json'
        response = api.rest(url, data, 'POST')
        new_user = json.loads(response)
        self.refresh()
        if 'id' in new_user:
            return int(new_user['id'])
        Utils.warning('Trying to create user (' + login_name + '), but ' +
                      'got an unexpected return from Skytap. Response:\n' +
                      response)
        return 0

    def delete(self, user, transfer_user):
        """Delete a user.

        .. warning::
            This is unrecoverable. Use with caution.

        Args:
            user: The user to delete (:class:`~skytap.models.User` or
                :class:`int`).
            transfer_user : Transfer all assets to this user
                (:class:`~skytap.models.User` or :class:`int`).

        Returns:
            bool: True if user deleted.

        Raises:
            TypeError: if user or transfer_user is not a
                :class:`~skytap.models.User` or :class:`int`.
            KeyError: If user or transfer_user isn't a user in the Users list.
        """
        if isinstance(user, int):
            if user not in self.data:
                raise KeyError
            user = self.data[user]
        if not isinstance(user, User):
            raise KeyError

        if isinstance(transfer_user, int):
            if transfer_user not in self.data:
                raise KeyError
            transfer_user = self.data[transfer_user]
        if not isinstance(transfer_user, User):
            raise KeyError

        target_id = user.id

        user.delete(transfer_user)
        self.refresh()

        return target_id not in self.data

if __name__ == '__main__':
    print(Users().main(sys.argv[1:]))



"""Support for an Template resource in Skytap."""
import json

from skytap.framework.ApiClient import ApiClient
from skytap.models.Notes import Notes
from skytap.models.SkytapResource import SkytapResource
from skytap.models.UserData import UserData
from skytap.models.Vms import Vms


class Template(SkytapResource):

    """One Skytap template."""

    def __init__(self, tmp_json):
        """Init is mainly handled by the parent class."""
        super(Template, self).__init__(tmp_json)

    def _calculate_custom_data(self):
        """Add custom data.

        Convert the list of VMs into a Vms object group.
        """
        self.data['vms'] = Vms(self.vms, self.id)

    def __getattr__(self, key):
        """Load values for anything that doesn't get loaded by default.

        For user_data, a secondary API call is needed. Only make that
        call when the info is requested.
        """
        if key == 'user_data':
            if key in self.data:
                return self.data[key]
            api = ApiClient()
            user_json = api.rest(self.url + '/user_data.json')
            self.user_data = UserData(json.loads(user_json))
            return self.user_data

        return super(Template, self).__getattr__(key)



"""Setup information for the Skytap API package."""
from setuptools import find_packages
from setuptools import setup

setup(
    name='skytap',
    packages=find_packages(),
    version='1.3.1',
    description='Skytap REST API access modules',
    author='Fulcrum Technologies',
    test_suite='nose.collector',
    author_email='bwellington@fulcrum.net',
    maintainer='Michael Knowles',
    maintainer_email='mknowles@fulcrum.net',
    license='MIT',
    install_requires=['requests', 'six'],
    scripts=['bin/skytap'],
    url='https://github.com/FulcrumIT/skytap',
    download_url='https://github.com/FulcrumIT/skytap/tarball/v1.2.0',
    keywords=['skytap', 'cloud', 'client', 'rest', 'api', 'development'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Systems Administration",
        "Topic :: Internet"
    ]
)



"""Support for Projects."""
from skytap.models.SkytapResource import SkytapResource
from skytap.Users import Users


class Project(SkytapResource):

    """One Skytap project."""

    def __init__(self, project_json):
        """Build one Skytap project."""
        super(Project, self).__init__(project_json)

    def _calculate_custom_data(self):
        """Make the list of users into a Users list."""
        self.data['users'] = Users(self.users)
        for key in self.users.keys():
            if self.users[key].url == self.owner_url:
                self.owner_name = self.users[key].name
                break



"""Some basic utility functions that help out with the Skytap API."""

from datetime import datetime
from datetime import timedelta
from datetime import tzinfo
import json
import logging


try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

# Set up the logging system:

logging.getLogger(__name__).addHandler(NullHandler())
logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')  # noqa
handler.setFormatter(formatter)
logger.addHandler(handler)


def debug(msg):
    logger.debug(msg)


def warning(msg):
    logger.warning(msg)


def info(msg):
    logger.info(msg)


def critical(msg):
    logger.critical(msg)


def log(level, msg):
    logger.log(level, msg)


def log_level(level=None):
    if level is not None:
        logger.setLevel(level)
    return logger.getEffectiveLevel()


def error(err):
    """Convert an error message into JSON."""
    logger.error(err)
    return json.dumps({"error": err})


def convert_date(date_str):
    """Convert a Skytap date string to a datetime object.

    Sample from Skytap: 2015/09/08 00:26:48 -0800
    """
    naive_date_str, _, offset_str = date_str.rpartition(' ')
    naive_dt = datetime.strptime(naive_date_str, '%Y/%m/%d %H:%M:%S')
    offset = int(offset_str[-4:-2]) * 60 + int(offset_str[-2:])
    if offset_str[0] == "-":
        offset = -offset
    dt = naive_dt.replace(tzinfo=FixedOffset(offset))
    return dt


class FixedOffset(tzinfo):

    """Fixed offset in minutes: `time = utc_time + utc_offset`."""

    def __init__(self, offset):
        """Given offset, built timezone info."""
        self.__offset = timedelta(minutes=offset)
        hours, minutes = divmod(offset, 60)
        # NOTE: the last part is to remind about deprecated POSIX GMT+h
        #  timezones that have the opposite sign in the name;
        #  the corresponding numeric value is not used e.g., no minutes
        self.__name = '<%+03d%02d>%+d' % (hours, minutes, -hours)

    def utcoffset(self, dt=None):
        """Return UTC offset."""
        return self.__offset

    def tzname(self, dt=None):
        """Return timezone name."""
        return self.__name

    def dst(self, dt=None):
        """Ignore daylight savings time."""
        return timedelta(0)

    def __repr__(self):
        """String representation of the offset."""
        total_seconds = (self.utcoffset().microseconds + 0.0 +
                         (self.utcoffset().seconds + self.utcoffset().days *
                         24 * 3600) * 10 ** 6) / 10 ** 6
        return 'FixedOffset(%d)' % (total_seconds / 60)



import json

from skytap.framework.ApiClient import ApiClient
import skytap.framework.Utils as Utils
from skytap.models.SkytapResource import SkytapResource


class UserData(SkytapResource):
    def __init__(self, contents, env_url):
        super(UserData, self).__init__(contents)
        self.url = env_url + '/user_data.json'

    def __str__(self):
        return self.contents

    def add(self, key, value):
        """Add value to environment's userdata.

        Args:
            key (str): The name of the value's key.
            value (str): The value to add.

        Returns:
            str: The response from Skytap, or "{}".
        """

        add_key = True

        lines = self.contents.split("\n")

        for i in lines:
            if i != "":
                j = i.split()
                if len(j) > 0 and j[0] == (key + ":"):
                    add_key = False

        if add_key:
            Utils.info('Adding key \"' + key + '\" with value \"'
                       '' + value + '\"')
            api = ApiClient()
            new_content = "" + key + ": " + value + "\n" + self.contents
            data = {"contents": new_content}
            response = api.rest(self.url, data, 'POST')
            self.data[key] = value
            self.refresh()
            return response
        else:
            Utils.info('Key \"' + key + '\" with value \"' + value + '\"'
                       'already exists.')
            return "{}"

    def delete(self, key):
        """Delete key/value from environment's userdata.

        Args:
            key (str): The name of key to delete, along with value

        Returns:
            str: The response from Skytap, or "{}".
        """

        new_content = ""

        del_key = False

        lines = self.contents.split("\n")

        for i in lines:
            if i != "":
                j = i.split()
                if len(j) > 0 and j[0] == (key + ":"):
                    del_key = True
                else:
                    new_content += (i.strip() + "\n")

        if del_key:
            Utils.info('Deleting key \"' + key + '\".')
            api = ApiClient()
            data = {"contents": "" + new_content}
            response = api.rest(self.url, data, 'POST')
            self.refresh()
            return response
        else:
            Utils.info('Key \"' + key + '\" already exists.')
            return "{}"

    def add_line(self, text, line=-1):
        """Add line to environment's userdata.

        Args:
            text (str): line of text to be added. (Required)
            line (int): line number to add to. If too large, default to last.

        Returns:
            str: The response from Skytap.
        """

        try:
            line = int(line)
        except ValueError:
            return "{}"  # Not an integer

        lines = self.contents.split("\n")

        new_content = ""

        line_found = False
        count = 0
        for i in lines:
            if i != "":
                if line == count:
                    new_content += (text.strip() + "\n")

                    new_content += (i.strip() + "\n")

                    line_found = True
                else:
                    new_content += (i.strip() + "\n")

            count += 1

        if not line_found:
            new_content += (text.strip() + "\n")

        Utils.info('Adding line: \"' + text + '\"')
        api = ApiClient()
        data = {"contents": new_content}
        response = api.rest(self.url, data, 'POST')
        self.refresh()
        return response

    def delete_line(self, line):
        """Delete line from environment's userdata.

        Args:
            line (int): line number to delete.

        Returns:
            str: The response from Skytap.
        """

        line = str(line)

        lines = self.contents.split("\n")

        new_content = ""

        line_found = False
        for i in lines:
            if i != "":
                if i.strip() != line.strip():
                        new_content += (i.strip() + "\n")
                        line_found = True

        Utils.info('Removing line: \"' + str(line) + '\"')
        api = ApiClient()
        data = {"contents": new_content.lstrip()}
        response = api.rest(self.url, data, 'POST')
        self.refresh()
        return response

    def get_line(self, line):
        """Return content of line from environment's userdata.

        Args:
            line (int): line number to get.

        Returns:
            str: The content of the line, or "".
        """

        try:
            line = int(line)
        except ValueError:
            return ""  # Not an integer

        lines = self.contents.split("\n")

        line_found = False
        count = 0
        for i in lines:
            if line == count:
                    return i

        if not line_found:
            return ""

    def _get_values(self, contents):
        """Check userdata and set variables based on keys/values within."""
        lines = contents.split("\n")

        values = {}

        for i in lines:
            tokens = i.split()

            if len(tokens) < 2:
                continue

            # Check for valid YAML formatting in first and second tokens in
            # each line, then add those values to dict.
            if (tokens[0].endswith(":") and "#" not in tokens[0] and
                    len(tokens) > 1 and "#" not in tokens[1]):
                    # If variable is a number, make it integer
                    try:
                        values[tokens[0][:-1]] = int(tokens[1])
                    except ValueError:
                        values[tokens[0][:-1]] = tokens[1]

                    self.data[tokens[0][:-1]] = values[tokens[0][:-1]]

        return values

    def _calculate_custom_data(self):
        if self.contents:
            values = self._get_values(self.contents)
        else:
            self.data["contents"] = ""



"""Support for an Environment resource in Skytap.

In nearly every case, you'll access an Environment via the
:class:`~skytap.Environments` object::

    envs = skytap.Environments()
    for environment in envs:
        print (environment.name)

You can access anything from an environment that Skytap includes in their
`API <http://help.skytap.com/#API_v2_Documentation.html#Environm>`_.
Most of these can be access directly as attributes of the given
Environment object::

    environment = skytap.Environments()[12345]
    print(environment.name)
    print(environment.json)

Some data conversions are handled for you. Specifically:

* Dates are converted into :class:`datetime` objects, like :attr:`created_at`.
* The :attr:`vms` list is loaded into a :class:`skytap.models.Vms` class.
* The :attr:`notes`, if loaded, are put into a :class:`skytap.models.Notes` class.
* The :attr:`user_data`, if loaded, is put into a :class:`skytap.models.UserData` class.


There's also the ability to change the runstate of the environment through
the function :func:`change_state`::

    environment = skytap.Environments()[12345]
    environment.change_state('suspended')

    # Passing `True` will wait for the suspend to complete
    # before returning to the script:
    environment.change_state('suspended', True)

The various state change options also have easy aliases available to them:

* :func:`run()`
* :func:`halt()`
* :func:`suspend()`
* :func:`reset()`
* :func:`stop()`

Passing `True` to any of these will also cause the script to wait until
the action is completed by Skytap.

.. note::
    Some pieces of a given environment, specifically `notes` and `user_data`,
    are only available via additional calls to the API. These fields will
    not exist when first creating the environments object, but any direct
    access to those fields will trigger the API call behind the scenes.

    This is important if you're listing the entire contents (say, sending it
    to a JSON) - these fields won't be included if you haven't made that direct
    access.

    This is by design to conserve API calls as most usage doesn't need or use
    those fields.
"""
import json

from skytap.framework.ApiClient import ApiClient
from skytap.framework.Suspendable import Suspendable
import skytap.framework.Utils as Utils
from skytap.models.Notes import Notes
from skytap.models.SkytapResource import SkytapResource
from skytap.models.UserData import UserData
from skytap.models.Vms import Vms


class Environment(SkytapResource, Suspendable):

    """One Skytap environment."""

    def __init__(self, env_json):
        """Init is mainly handled by the parent class."""
        super(Environment, self).__init__(env_json)

    def _calculate_custom_data(self):
        """Add custom data.

        Specifically, boolean values to more easily determine state, allowing
        things like 'if env.running:' to be used.
        """
        self.data['running'] = self.runstate == 'running'
        self.data['busy'] = self.runstate == 'busy'
        self.data['suspended'] = self.runstate == 'suspended'
        self.data['vms'] = Vms(self.vms, self.id)

    def __getattr__(self, key):
        """Load values for anything that doesn't get loaded by default.

        For user_data and notes, a secondary API call is needed. Only make that
        call when the info is requested.
        """
        if key == 'user_data':
            if key in self.data:
                return self.data[key]
            api = ApiClient()
            user_json = api.rest(self.url + '/user_data.json')
            self.data['user_data'] = UserData(json.loads(user_json), self.url)
            return self.user_data

        if key == 'notes':
            api = ApiClient()
            notes_json = api.rest(self.url + '/notes.json')
            self.notes = Notes(notes_json, self.url)
            return self.notes

        return super(Environment, self).__getattr__(key)

    def delete(self):
        """Delete the environment.

        In general, it'd seem wise not to do this very often.
        """
        Utils.info('Deleting environment: ' +
                   str(self.id) + '(' + self.name + ')')
        api = ApiClient()
        response = api.rest(self.url_v1,
                            {},
                            'DELETE')
        return response



import json

from skytap.models.SkytapResource import SkytapResource


class Vpn(SkytapResource):
    def __init__(self, vpn_json):
        super(Vpn, self).__init__(vpn_json)

    def _calculate_custom_data(self):
        self.active = self.status == 'active'



"""Support for a VM resource in Skytap."""
import json

from skytap.framework.ApiClient import ApiClient
from skytap.framework.Suspendable import Suspendable
import skytap.framework.Utils as Utils
from skytap.models.Interfaces import Interfaces
from skytap.models.Notes import Notes
from skytap.models.SkytapResource import SkytapResource
from skytap.models.UserData import UserData


class Vm(SkytapResource, Suspendable):

    """One Skytap VM."""

    def __init__(self, vm_json):
        """Init is mainly handled by the parent class."""
        super(Vm, self).__init__(vm_json)

    def _calculate_custom_data(self):
        """Add custom data.

        Specifically, boolean values to more easily determine state, allowing
        things like 'if vm.running:' to be used.
        """
        self.data['running'] = self.runstate == 'running'
        self.data['busy'] = self.runstate == 'busy'
        self.data['suspended'] = self.runstate == 'suspended'

    def __getattr__(self, key):
        """Load values for anything that doesn't get loaded by default.

        For user_data, notes, and interfaces, a secondary API call is needed.
        Only make that call when the info is requested.
        """
        if key == 'user_data':
            if key in self.data:
                return self.data[key]
            api = ApiClient()
            user_json = api.rest(self.url + '/user_data.json')
            self.user_data = UserData(json.loads(user_json), self.url)
            return self.user_data

        if key == 'notes':
            api = ApiClient()
            notes_json = api.rest(self.url + '/notes.json')
            self.notes = Notes(notes_json, self.url)
            return self.notes

        if key == 'interfaces':
            if key in self.data:
                return self.data[key]
            api = ApiClient()
            interfaces_json = json.loads(api.rest(self.url))
            self.interfaces = Interfaces(interfaces_json["interfaces"], self.url)
            return self.interfaces

        return super(Vm, self).__getattr__(key)

    def delete(self):
        """Delete a VM.

        In general, it'd seem wise not to do this very often.
        """
        Utils.info('Deleting VM: ' + str(self.id) + '(' + self.name + ')')
        api = ApiClient()
        response = api.rest(self.url,
                            {},
                            'DELETE')
        return response
